/* risc.cpp: RISC Compiler (c) 2015 CECS (TS, GL, RD) */

#include <iostream>

#include "risc.h"

#define HELP_PAGE                                                                   \
"risc: "RISC_LIBRARY_NAME" Compiler, V"RISC_VERSION_NUMBER"\n"                      \
"      "RISC_COPYRIGHT"\n"                                                          \
"\n"                                                                                \
"Usage: risc [<option>...] <design> [<option>...]\n"                                \
"\n"                                                                                \
"Options:               Description:                    Default:\n"                 \
"-h|--help              print this usage info, quit     none\n"                     \
"-v|--verbose           increment verbosity             silent\n"                   \
"-vv                    increment verbosity by 2        silent\n"                   \
"-vvv                   increment verbosity by 3        silent\n"                   \
"-w|--warnings          increment warning level         no warnings\n"              \
"-ww                    increment warning level by 2    no warnings\n"              \
"-www                   increment warning level by 3    no warnings\n"              \
"-g                     produce debugging information   no debugging infos\n"       \
"-O|-O<level>           enable optimizations            no optimization\n"          \
"-I<dir>                add dir to include path         $SYSTEMC_LW_HOME/include\n" \
"-L<dir>                add dir to library path         $SYSTEMC_OOP_HOME/lib\n"    \
"-l<lib>                add lib to linker libraries     -lsystemc\n"                \
"-c                     compile only, do not link       compile and link\n"         \
"-o <filename>          name output file as filename    a.out\n"                    \
"-risc:dump             output SG, conflict tables      none\n"                     \
"-risc:icpc             use Intel compiler backend      g++\n"                      \
"-risc:mic              compile for MIC architecture    none\n"                     \
"-risc:elab <filename>  import elaboration result       none\n"                     \
"<rose:option>          passed to ROSE compiler         none\n"                     \
"<g++_option>           passed to GNU C++ compiler      none\n"                     \
"\n"

class RISCParseStatus: public risc::tools::ParseStatus
{
  public:
    bool dump;

    RISCParseStatus(): ParseStatus(), dump(false) {}
    RISCParseStatus& operator=(const RISCParseStatus& copy)
    {
      verbosityLevel = copy.verbosityLevel;
      warningLevel = copy.warningLevel;
      compilerName = copy.compilerName;
      mic = copy.mic;
      dump = copy.dump;

      return *this; 
    }
};

void
RISCParseCommandlineOptions(int &argc, char** &argv, RISCParseStatus* status)
{
  if (argc < 2)
  {
    std::cout << "risc: no input files\n";
    exit(0);
  }

  Rose_STL_Container<std::string> argList;
  argList = CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

  std::vector<Rose_STL_Container<std::string>::iterator> RemovedIter;

  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    size_t pos;

    // -h, --help
    if (it->compare("-h") == 0 || it->compare("--help") == 0)
    {
      std::cout << HELP_PAGE;
      exit(0);
    }

    // -risc:icpc
    if (it->compare("-risc:icpc") == 0)
    {
      status->compilerName = "icpc";
      RemovedIter.push_back(it);
    }

    // -risc:mic
    if (it->compare("-risc:mic") == 0)
    {
      status->mic = true;
      RemovedIter.push_back(it);
    }

    // -risc:dump
    if (it->compare("-risc:dump") == 0)
    {
      status->dump = true;
      RemovedIter.push_back(it);
    }
  }

  while (RemovedIter.size() != 0)
  {
    argList.erase(RemovedIter.back());
    RemovedIter.pop_back();
  }

  argc = 0;
  argv = NULL;

  CommandlineProcessing::generateArgcArgvFromList(argList, argc, argv);

  // Parse library-level command-line options
  risc::tools::ParseCommandlineOptions(argc, argv, status);
}

int main(int argc, char* argv[])
{
  RISCParseStatus status;

  RISCParseCommandlineOptions(argc, argv, &status);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  { 
    std::cout << "RISC Compiler starting..."
              << std::endl;
  }

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "RISC: adding command-line options for frontend..."
              << std::endl;
  }

  std::string LWIncludePath;
  std::string LWLibraryPath;

  LWIncludePath = "-I";
  LWIncludePath += getenv(SYSTEMC_LW_HOME_ENV);
  LWIncludePath += "/include";
  LWLibraryPath = "-L";
  LWLibraryPath += getenv(SYSTEMC_LW_HOME_ENV);
  LWLibraryPath += "/lib-linux64";

  // add -I$SYSTEMC_LW_HOME/include
  risc::tools::AddCommandlineOption(argc, argv, LWIncludePath);

  // add -L$SYSTEMC_LW_HOME/lib-linux64
  risc::tools::AddCommandlineOption(argc, argv, LWLibraryPath);

  // add -lsystemc
  risc::tools::AddCommandlineOption(argc, argv, "-lsystemc");

  // add -mmic
  if (status.mic == true)
  {
    risc::tools::AddCommandlineOption(argc, argv, "-mmic");
  }

  if (status.verbosityLevel >= RISC_VERBOSE)
  { 
    std::cout << "RISC: parsing, building ROSE IR..."
              << std::endl;
  }
  risc::Design d(argc,argv);

  if (status.verbosityLevel >= RISC_VERBOSE)
  { 
    std::cout << "RISC: analyzing, building RISC SG..."
              << std::endl;
  }
  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  { 
    std::cout << "RISC: listing model hierarchy..."
              << std::endl;
  }
  for (risc::ModuleVectorIter
       iter_module  = d.module_definitions_.begin();
       iter_module != d.module_definitions_.end();
       iter_module++) {
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    { 
      std::cout << "  Module: " <<  (*iter_module)->get_name() << std::endl;
    }
    for (risc::ThreadVectorIter
         iter_thread = (*iter_module)->threads_.begin();
         iter_thread != (*iter_module)->threads_.end();
         iter_thread++) {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      { 
        std::cout << "    " << (*iter_thread)->get_name() << std::endl;
      }
      sg.add_function(*iter_thread);
    }
  }

  sg.clean_graph();

  if (status.verbosityLevel >= RISC_VERBOSE)
  { 
    std::cout << "RISC: analyzing, building conflict tables..."
              << std::endl;
  }
  // Time advance tables
  risc::sg::TimeAdvanceTable time_advance_table(sg);
  if (status.dump == true)
  {
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    { 
      std::cout << "RISC: generating \"current_time_advance_table.html\"..."
                << std::endl;
    }
    time_advance_table.print_current_time_advance_table(
                          "current_time_advance_table.html");
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << "RISC: generating \"next_time_advance_table.html\"..."
                << std::endl;
    }
    time_advance_table.print_next_time_advance_table(
                          "next_time_advance_table.html");
  }
  // Event conflict table
  risc::sg::EventConflictTable
    event_conflict_table(sg, d.path_instance_mapper_, true);
  if (status.dump == true)
  {
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    { 
      std::cout << "RISC: generating \"event_conflict_table.html\"..."
                << std::endl;
    }
    event_conflict_table.print_conflict_table("event_conflict_table.html");

    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << "RISC: generating \"conflict_lookup_table.html\"..."
                << std::endl;
    }
    event_conflict_table.print_lookup_table("conflict_lookup_table.html");
  }
  // Data conflict table
  risc::sg::DataConflictTable
    data_conflict_table(sg, d.path_instance_mapper_, true);
  if (status.dump == true)
  {
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    { 
      std::cout << "RISC: generating \"data_conflict_table.html\"..."
                << std::endl;
    }
    data_conflict_table.print_conflict_table("data_conflict_table.html");

    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    { 
      std::cout << "RISC: generating \"segment_graph.dot\"..."
                << std::endl;
    }
    sg.print_graph("segment_graph.dot");
  }

  if (status.verbosityLevel >= RISC_VERBOSE)
  { 
    std::cout << "RISC: instrumenting, generating parallel model..."
              << std::endl;
  }

  risc::inst::instrumentor(d, sg, &data_conflict_table, &event_conflict_table, 
                           &time_advance_table, &status);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "RISC: modifying command-line options for backend..."
              << std::endl;
  }

  // Change the command line argument list

  // remove -I$SYSTEMC_LW_HOME/include
  risc::tools::RemoveCommandlineOption(d, LWIncludePath);

  // remove -L$SYSTEMC_LW_HOME/lib-linux64
  risc::tools::RemoveCommandlineOption(d, LWLibraryPath);

  if (status.mic == true)
  {
    std::string MICIncludePath;
    std::string MICLibraryPath;

    MICIncludePath = "-I";
    MICIncludePath += getenv(SYSTEMC_MIC_HOME_ENV);
    MICIncludePath += "/include";
    MICLibraryPath = "-L";
    MICLibraryPath += getenv(SYSTEMC_MIC_HOME_ENV);
    MICLibraryPath += "/lib-linux64";

    // add -I$SYSTEMC_MIC_HOME/include
    risc::tools::AddCommandlineOption(d, MICIncludePath);

    // add -L$SYSTEMC_MIC_HOME/lib-linux64
    risc::tools::AddCommandlineOption(d, MICLibraryPath);
  } else {
    std::string OOPIncludePath;
    std::string OOPLibraryPath;
    std::string OOPLibraryPathwoL;

    OOPIncludePath = "-I";
    OOPIncludePath += getenv(SYSTEMC_OOP_HOME_ENV);
    OOPIncludePath += "/include";
    OOPLibraryPath = "-L";
    OOPLibraryPath += getenv(SYSTEMC_OOP_HOME_ENV);
    OOPLibraryPath += "/lib-linux64";
    OOPLibraryPathwoL = OOPLibraryPath;
    OOPLibraryPathwoL.erase(0, 2);

    // add -I$SYSTEMC_OOP_HOME/include
    risc::tools::AddCommandlineOption(d, OOPIncludePath);

    // add -L$SYSTEMC_OOP_HOME/lib-linux64
    risc::tools::AddCommandlineOption(d, OOPLibraryPath);

    // add -Xlinker -R -Xlinker $SYSTEMC_OOP_HOME/lib-linux64
    risc::tools::AddCommandlineOption(d, "-Xlinker");
    risc::tools::AddCommandlineOption(d, "-R");
    risc::tools::AddCommandlineOption(d, "-Xlinker");
    risc::tools::AddCommandlineOption(d, OOPLibraryPathwoL);
  }

  // risc backend
  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "RISC: unparsing source files..." << std::endl;
  }

  risc::inst::risc_unparse(d, &status, NULL, NULL);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "RISC: patching ROSE limitations..."
              << std::endl;
  }

  risc::inst::risc_patch(d, &status, 
      "sed -r -i \'s/create_method_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_method_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5, \\6, \\7)/\' "RISC_UNPARSE_FILE_PREFIX"%s",
      "sed -r -i \'s/create_thread_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_thread_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5, \\6, \\7)/\' "RISC_UNPARSE_FILE_PREFIX"%s",
      "sed -r -i \'s/create_cthread_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_cthread_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5, \\6, \\7)/\' "RISC_UNPARSE_FILE_PREFIX"%s");


  if (status.verbosityLevel >= RISC_VERBOSE)
  {
    std::cout << "RISC: compiling executable..."
              << std::endl;
  }

  risc::inst::risc_compile(d, &status);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  { 
    std::cout << "RISC Compiler done."
              << std::endl;
  }
  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
