#include "event_conflict_table.h"

#include <list>

#include "mapped_variable.h"
#include "segment.h"
#include "segment_graph.h"
#include "../tools/class_is_derived_from.h"
#include "../tools/misc.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/instance.h"
#include "../internal_representation/internal_representation_helper.h"
#include "../internal_representation/module.h"
#include "../internal_representation/path_instance_mapper.h"

std::list<risc::sg::MappedVariable>
risc::sg::EventConflictTable::get_all_list_events_dynamically(
    SgVariableSymbol *sc_event_list, int instance_id)
{
  // here we are getting the variable of the event list.
  // Just in case that the event list is a reference
  MappedVariableList mapped_event_list
    = get_mapped_variable_dynamically(sc_event_list, instance_id);

  // getting the variable declaration from the event list
  SgVariableDeclaration *variable
    = isSgVariableDeclaration(
          mapped_event_list.front().symbol_->get_declaration()
                                               ->get_declaration());

  SgClassDefinition *enclosing_class_def_of_event_list
    = SageInterface::getEnclosingClassDefinition(variable);

  // getting the name of the variable
  std::string host_module_type_name
    = risc::tools::get_mangled_name(enclosing_class_def_of_event_list);
  std::string event_list_name = SageInterface::get_name(variable);
  event_list_name.erase(0, 22);

  // here we are getting the DynamicObject of the event list in the tree
  int counter = 0;
  DynamicObject *mapped_variable_address
    = get_dynamic_object_of_variable(
          path_instance_mapper_->get_dynamic_tree_root(),
          host_module_type_name,
          mapped_event_list.front().instance_id_,
          event_list_name,
          counter);

  DynamicEventList *dynamic_event_list
    = dynamic_cast<DynamicEventList*>(mapped_variable_address);

  std::list<risc::sg::MappedVariable> list_events;

  for(std::list<int*>::iterator
      iter  = dynamic_event_list->events_.begin();
      iter != dynamic_event_list->events_.end();
      iter++) {

    // Searching for the host module and the variable name of the variable
    // address
    std::pair<DynamicModule*, std::string> host_module_and_variable
      = get_host_module_and_name_of_variable(
          path_instance_mapper_->get_dynamic_tree_root(),
          *iter);

    // If we are reaching this point, it is a member variable
    // So, next we determine the instance id of the hosting module of the
    // variable
    int mapped_instance_id = 0;

    bool result = get_instance_id_of_hosting_module(
                      path_instance_mapper_->get_dynamic_tree_root(),
                      *iter,
                      host_module_and_variable.first->name_,
                      mapped_instance_id);

    // At this point, we have to find the associated module definition
    // in the internal representation. We can do this by comparing the
    // modules types (mangled names)
    risc::Module *current_module
      = risc::find_module_definition(host_module_and_variable.first->name_);

    SgVariableSymbol *mapped_symbol = NULL;

    // After having the module, we have to search for the mapped variable in the
    // class. So, we are iterating over all the declarations in the class and
    // compare all the variables names with name of the mapped variable
    for(SgDeclarationStatementPtrList::iterator
        member_iter  = current_module->get_ast_node()->get_members().begin();
        member_iter != current_module->get_ast_node()->get_members().end();
        member_iter++) {

      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
      if(var_decl) {

        std::string temp_variable_name(SageInterface::get_name(var_decl));

        // deleting __variable_declaration__ out of the string
        temp_variable_name.erase(0, 22);

        if(temp_variable_name == host_module_and_variable.second) {
          mapped_symbol = isSgVariableSymbol(var_decl->get_decl_item(0)
              ->search_for_symbol_from_symbol_table());
        }
      }
    }

    // pushing the mapped event into the event list
    list_events.push_back(MappedVariable(mapped_symbol, mapped_instance_id));
  }

  return list_events;
}

std::list<risc::sg::MappedVariable>
risc::sg::EventConflictTable::get_all_list_events_statically(
    SgSymbol *sc_event_list, int instance_id)
{
  std::list<MappedVariable> result;

  SgClassDefinition *class_def
    = SageInterface::getEnclosingClassDefinition(sc_event_list);

  std::vector<SgNode*> func_defs
    = NodeQuery::querySubTree(class_def, V_SgFunctionDefinition);

  SgFunctionDefinition *before_end_of_elaboration = NULL;

  for(std::vector<SgNode*>::iterator
      iter  = func_defs.begin();
      iter != func_defs.end();
      iter++) {

    SgFunctionDefinition *func_def = isSgFunctionDefinition(*iter);
    if(SageInterface::get_name(func_def) == "before_end_of_elaboration") {
      before_end_of_elaboration = func_def;
    }
  }

  for(SgStatementPtrList::iterator
      iter  = before_end_of_elaboration->get_body()->get_statements().begin();
      iter != before_end_of_elaboration->get_body()->get_statements().end();
      iter++) {

    SgExprStatement *expr_stmt = isSgExprStatement(*iter);

    if(!expr_stmt) {
      continue;
    }

    SgExpression *expr = isSgExpression(expr_stmt->get_expression());

    if(!expr) {
      continue;
    }

    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expr);
    if(func_call_exp) {

      if((SageInterface::get_name(func_call_exp).find("&=")) ||
          (SageInterface::get_name(func_call_exp).find("|="))) {

        SgVarRefExp *assigned_var =
          isSgVarRefExp(
              isSgArrowExp(
                  isSgDotExp(
                      func_call_exp->get_function())
                  ->get_lhs_operand())
              ->get_rhs_operand());

        if(assigned_var) {

          if(assigned_var->get_symbol() == sc_event_list) {

            SgExprListExp *expr_list = func_call_exp->get_args();

            SgVarRefExp *var
              = isSgVarRefExp(
                    isSgArrowExp(
                        expr_list->get_expressions().at(0))
                    ->get_rhs_operand());

            MappedVariableList mapped_event
              = get_mapped_variable_statically(var->get_symbol(), instance_id);

            result.push_back(mapped_event.front());
          }
        }
      }
    }
  }

  return result;
}

risc::sg::EventConflictTable::EventConflictTable(
    SegmentGraph &graph,
    PathInstanceMapper *path_instance_mapper,
    bool debugging):
  ConflictTable(graph, path_instance_mapper, debugging)
{
  determine_conflict_table();
}

void
risc::sg::EventConflictTable::determine_conflict_table()
{
  bool dynamic_analysis = path_instance_mapper_->perform_dynamic_analysis();

  for(int column=0; column < size_of_conflict_table_; column++) {

    // Here, we are getting the segment id and instance id with the associated
    // column index 'column'.
    std::pair<int, int> column_id_and_instance =
      index_to_segment_and_instance_id(column);

    // Next, we have to get the boundary statement
    const SgNode *boundary
      = segment_graph_.graph_[
            segment_id_to_vertex_descriptor(column_id_and_instance.first)].
                boundary_node_;

    const SgStatement *stmt = isSgStatement(boundary);
    const SgExprStatement *expr_stmt = isSgExprStatement(stmt);


    std::list<MappedVariable> waiting_events;

    // Here, we have to check what kind of boundary statement it is. If it is a
    // wait function call, we have to analyze the arguments. So, if we have
    // exactly one argument it could be an event or an event list. We will check
    // if the symbol is of the type event.
    if(expr_stmt) {

      const SgFunctionCallExp *func_call
        = isSgFunctionCallExp(expr_stmt->get_expression());

      if(func_call) {

        if(SageInterface::get_name(
               func_call->getAssociatedFunctionDeclaration()) == "wait") {

          // At this point, we know that the boundary statement is a function
          // call of wait.
          // Next we have to figure out what are the arguments

          // In case we have an immediate notification.
          // We can not support this one
          if(func_call->get_args()->get_expressions().size() == 0) {

            assert(0 && "EventConflictTable::determine_conflict_table:" &&
                " No support of immediate notification");

          } else if(func_call->get_args()->get_expressions().size() == 1) {

            // A sc_zero_time will wait for one delta count. We don't have to
            // set the wait symbol
            if(func_call->get_args()->get_expressions().at(0)->unparseToString()
                == "SC_ZERO_TIME") {

            } else {

              // Here we are getting the variable for which the wait function
              // call is waiting
              SgVarRefExp* waiting_event
                = isSgVarRefExp(isSgArrowExp(func_call->get_args()->
                      get_expressions().at(0))->get_rhs_operand());


              if(isSgPointerType(waiting_event->get_type())) {

                assert(0 && "Event pointers are not supported");

              } else if(isSgReferenceType(waiting_event->get_type())) {

                // If it is a reference type, we use the InstanceTree to get
                // the real mapped variable
                MappedVariableList mapped_event;
                if(dynamic_analysis) {
                  mapped_event = get_mapped_variable_dynamically(
                                     waiting_event->get_symbol(),
                                     column_id_and_instance.second);
                } else {
                  mapped_event = get_mapped_variable_statically(
                                     waiting_event->get_symbol(),
                                     column_id_and_instance.second);
                }

                waiting_events.push_back(mapped_event.front());

              } else if(risc::tools::is_sc_event(waiting_event)) {

                // If it a event variable, we can set the symbol straight
                MappedVariable temp(waiting_event->get_symbol(),
                                   column_id_and_instance.second);
                waiting_events.push_back(temp);

              } else {
                // here we are covering the case of an event list

                if(dynamic_analysis) {

                  waiting_events = get_all_list_events_dynamically(
                                       waiting_event->get_symbol(),
                                       column_id_and_instance.second);
                } else {

                  waiting_events = get_all_list_events_statically(
                                       waiting_event->get_symbol(),
                                       column_id_and_instance.second);
                }
              }
            }
          }
        }
      }
    }

    // At this point, we have the waiting symbol of the wait function call.
    // We have to figure out, if this symbol is notified in each of the
    // segments.
    for(int row=0; row < size_of_conflict_table_; row++) {

      conflict_table_[row][column] = false;

      // Here, we are getting the segment id and instance id with the associated
      // row index 'row'.
      std::pair<int, int> row_id_and_instance =
        index_to_segment_and_instance_id(row);

      // We are iterating over all expressions for each segment. For an
      // individual segment we have to check if there are any nofitications of
      // events
      for(Segment::Expressions::iterator
          iter  = segment_graph_.graph_[
                      segment_id_to_vertex_descriptor(
                        row_id_and_instance.first)].expressions_.begin();
          iter != segment_graph_.graph_[
                      segment_id_to_vertex_descriptor(
                        row_id_and_instance.first)].expressions_.end();
          iter++) {

        SgExprStatement *expr_stmt = isSgExprStatement(*iter);
        SgFunctionCallExp *func_call = isSgFunctionCallExp(*iter);

        if(func_call) {

          SgFunctionDeclaration *func_decl
            = func_call->getAssociatedFunctionDeclaration();

          if(func_decl) {

            if(SageInterface::get_name(func_decl) == "notify") {

              // In case we have an immediate notification.
              // We can not support this one
              if(func_call->get_args()->get_expressions().size() == 0) {

                assert(0 && "EventConflictTable::determine_conflict_table:" &&
                    " No support of immediate notification");

              } else if(func_call->get_args()->get_expressions().size() == 1) {

                // At this point we know that it is a event notification
                SgExpression *lhs = isSgDotExp(
                                        func_call->get_function())->
                                            get_lhs_operand();

                SgVarRefExp *notify_event = isSgVarRefExp(isSgArrowExp(lhs)->
                                           get_rhs_operand());

                MappedVariableList notifying_events;
                if(isSgPointerType(notify_event->get_type())) {

                  assert(0 && "Event pointers are not supported");

                } else if(risc::tools::is_sc_event(notify_event)) {

                  // If it a event variable, we can set the symbol straight
                  MappedVariable notifying_event(
                      notify_event->get_symbol(), row_id_and_instance.second);
                  notifying_events.push_back(notifying_event);

                } else if(isSgReferenceType(notify_event->get_type())) {

                  if(dynamic_analysis) {
                    notifying_events = get_mapped_variable_dynamically(
                                           notify_event->get_symbol(),
                                           row_id_and_instance.second);
                  } else {
                    notifying_events = get_mapped_variable_statically(
                                           notify_event->get_symbol(),
                                           row_id_and_instance.second);
                  }
                }

                for(MappedVariableList::iterator
                    iter_notifying_event  = notifying_events.begin();
                    iter_notifying_event != notifying_events.end();
                    iter_notifying_event++) {

                  std::list<MappedVariable>::iterator conflict
                    = std::find(
                               waiting_events.begin(),
                               waiting_events.end(),
                               *iter_notifying_event);

                  if(conflict != waiting_events.end()) {

                    conflict_table_[row][column] = true;
                    if(debugging_) {
                      conflict_info_[row][column].conflict_variables_.insert(
                                                      *iter_notifying_event);
                    }
                  }
                }
              } else {
                assert(0 && "Unsuppported case");
              }
            }
          }
        }
      }
    }
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
