#include "segment.h"

#include "cached_function_ast_attributes.h"
#include "../internal_representation/design.h"
#include "../tools/function_helper.h"
#include "../tools/misc.h"


int risc::sg::Segment::id_counter = 0;

risc::sg::Segment::Segment(
    SgNode const * const boundary_node,
    bool inlined_segment):
  id_(id_counter),
  boundary_node_(boundary_node),
  in_conflict_with_all_other_segments(false),
  color_for_communication_graph_(Untouched),
  helper_for_graph_algorithms_(Untouched),
  inlined_segment_(inlined_segment)
{
  Segment::id_counter++;
  debugging_ = true;
}

risc::sg::Segment::Segment(const Segment &other):
  id_(other.id_),
  boundary_node_(other.boundary_node_),
  expressions_(other.expressions_),
  in_conflict_with_all_other_segments(false),
  color_for_communication_graph_(Untouched),
  helper_for_graph_algorithms_(Untouched)
{
  debugging_ = true;
}

bool
risc::sg::Segment::operator==(const Segment &other)
{
  return id_ == other.id_;
}

void
risc::sg::Segment::add_expression(SgNode* node)
{
  expressions_.push_back(node);
}

void
risc::sg::Segment::merge(const Segment &other)
{
  if(*this == other)
    return;

  in_conflict_with_all_other_segments |=
    this->in_conflict_with_all_other_segments;

  for(Expressions::const_iterator
      iter  = other.expressions_.begin();
      iter != other.expressions_.end();
      iter++) {
    add_expression(*iter);
  }
}

void
risc::sg::Segment::analyze_segment()
{
  for(Segment::Expressions::iterator
      iter  = expressions_.begin();
      iter != expressions_.end();
      iter++) {

    if(isSgExpression(*iter)) {
      SgExpression *expr = isSgExpression(*iter);

      if(isSgAssignOp((*iter)->get_parent())) {
        SgAssignOp *assign_op = isSgAssignOp((*iter)->get_parent());
        if(assign_op->get_lhs_operand() == *iter) {

          analyze_expression(expr, true, false, false, false, false, NULL);
          continue;
        }
      }

      analyze_expression(expr, false, false, false, false, false, NULL);
    }

    if(isSgExprStatement(*iter)) {
      SgExprStatement *stmt_expr = isSgExprStatement(*iter);
      analyze_expression(stmt_expr->get_expression(), false, false, false,
          false, false, NULL);
    }

    if(isSgVariableDeclaration(*iter)) {

      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);
      SgInitializedName *initialized_name = var_decl->get_decl_item(0);

      if(initialized_name->get_initptr()) {

        write_variables_.insert(
                             isSgVariableSymbol(
                                 initialized_name->
                                     search_for_symbol_from_symbol_table()));

        analyze_expression(initialized_name->get_initptr(),
            false, false, false, false, false, NULL);
      }
    }
  }
}


void
risc::sg::Segment::analyze_expression(SgNode *expression,
    bool left_of_assignment, bool read_write, bool member_access,
    bool function_parameter, bool is_precached_function,
    SgVariableSymbol *outer_variable)
{
  #define ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(Operator)              \
  if(is##Operator(expression)) {                                    \
    Operator *op = is##Operator(expression);                        \
    analyze_expression(op->get_lhs_operand(), true, false, false,   \
        function_parameter, is_precached_function, outer_variable); \
    analyze_expression(op->get_lhs_operand(), false, false, false,  \
        function_parameter, is_precached_function, outer_variable); \
    analyze_expression(op->get_rhs_operand(), false, false, false,  \
        function_parameter, is_precached_function, outer_variable); \
    return;                                                         \
  }
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgAndAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgIorAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgXorAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgLshiftAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgRshiftAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgPlusAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgMinusAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgMultAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgDivAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgModAssignOp);
  #undef ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP

  #define ANALYZE_VARS_IN_BINARY_OP(Operator)                       \
  if(is##Operator(expression)) {                                    \
    Operator *op = is##Operator(expression);                        \
    analyze_expression(op->get_lhs_operand(), false, false, false,  \
        function_parameter, is_precached_function, outer_variable); \
    analyze_expression(op->get_rhs_operand(), false, false, false,  \
        function_parameter, is_precached_function, outer_variable); \
    return;                                                         \
  }
  ANALYZE_VARS_IN_BINARY_OP(SgLessThanOp);
  ANALYZE_VARS_IN_BINARY_OP(SgGreaterThanOp);
  ANALYZE_VARS_IN_BINARY_OP(SgLessOrEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgGreaterOrEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgEqualityOp);
  ANALYZE_VARS_IN_BINARY_OP(SgNotEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgLshiftOp);
  ANALYZE_VARS_IN_BINARY_OP(SgRshiftOp);

  ANALYZE_VARS_IN_BINARY_OP(SgBitAndOp);
  ANALYZE_VARS_IN_BINARY_OP(SgBitOrOp);
  ANALYZE_VARS_IN_BINARY_OP(SgAndOp);
  ANALYZE_VARS_IN_BINARY_OP(SgOrOp);
  ANALYZE_VARS_IN_BINARY_OP(SgBitXorOp);

  ANALYZE_VARS_IN_BINARY_OP(SgAddOp);
  ANALYZE_VARS_IN_BINARY_OP(SgSubtractOp);
  ANALYZE_VARS_IN_BINARY_OP(SgDivideOp);
  ANALYZE_VARS_IN_BINARY_OP(SgMultiplyOp);
  ANALYZE_VARS_IN_BINARY_OP(SgModOp);

  ANALYZE_VARS_IN_BINARY_OP(SgCommaOpExp);
  #undef ANALYZE_VARS_IN_BINARY_OP


  #define ANALYZE_VARS_IN_UNARY_OP(Operator)                        \
  if(is##Operator(expression)) {                                    \
    Operator *op = is##Operator(expression);                        \
    analyze_expression(op->get_operand(), false, false, false,      \
        function_parameter, is_precached_function, outer_variable); \
    return;                                                         \
  }
  ANALYZE_VARS_IN_UNARY_OP(SgUnaryAddOp);
  ANALYZE_VARS_IN_UNARY_OP(SgMinusOp);
  ANALYZE_VARS_IN_UNARY_OP(SgAssignInitializer);
  ANALYZE_VARS_IN_UNARY_OP(SgNotOp);
  ANALYZE_VARS_IN_UNARY_OP(SgBitComplementOp);
  ANALYZE_VARS_IN_UNARY_OP(SgCastExp);
  #undef ANALYZE_VARS_IN_UNARY_OP

  #define ANALYZE_VARS_IN_UNARY_INC_DEC_OP(Operator)                \
  if(is##Operator(expression)) {                                    \
    Operator *op = is##Operator(expression);                        \
    analyze_expression(op->get_operand(), false, true, false,       \
        function_parameter, is_precached_function, outer_variable); \
    return;                                                         \
  }
  ANALYZE_VARS_IN_UNARY_INC_DEC_OP(SgPlusPlusOp);
  ANALYZE_VARS_IN_UNARY_INC_DEC_OP(SgMinusMinusOp);
  #undef ANALYZE_VARS_IN_UNARY_INC_DEC_OP

  #define ANALYZE_VARS_IN_ACCESS_OP(Operator)                         \
  if(is##Operator(expression)) {                                      \
    Operator *op = is##Operator(expression);                          \
    analyze_expression(op->get_rhs_operand(), left_of_assignment,     \
                           read_write, member_access,                 \
                           function_parameter, is_precached_function, \
                           outer_variable);                           \
    analyze_expression(op->get_lhs_operand(), left_of_assignment,     \
                           read_write, true, function_parameter,      \
                           is_precached_function, outer_variable);    \
    return;                                                           \
  }

  ANALYZE_VARS_IN_ACCESS_OP(SgArrowExp);
  ANALYZE_VARS_IN_ACCESS_OP(SgDotExp);

  ANALYZE_VARS_IN_ACCESS_OP(SgDotStarOp);
  ANALYZE_VARS_IN_ACCESS_OP(SgArrowStarOp);
  #undef ANALYZE_VARS_IN_ACCESS_OP

  if(isSgSizeOfOp(expression)) {
    SgSizeOfOp *size_of_op = isSgSizeOfOp(expression);
    analyze_expression(size_of_op->get_operand_expr(), false, false, false,
        function_parameter, is_precached_function, outer_variable);
  }

  if(isSgVarRefExp(expression)) {
    SgVarRefExp *var_ref = isSgVarRefExp(expression);

    if(!member_access) {

      if(function_parameter) {

        if(is_precached_function) {

          //write_variables_.insert(var_ref->get_symbol());
          //
          // This is the case where we have a function call like
          // foo(array[i+j]); We have only reading access to i and j, however,
          // we don't have any acces to array. The outer_variable represents
          // the symbol of array
          if(var_ref->get_symbol() != outer_variable) {

            read_variables_.insert(var_ref->get_symbol());

            if(read_write) {
              write_variables_.insert(var_ref->get_symbol());
            }

            if(left_of_assignment) {
              write_variables_.insert(var_ref->get_symbol());
            }
          }

        } else {
          write_variables_.insert(var_ref->get_symbol());
          read_variables_.insert(var_ref->get_symbol());
        }
        return;

      } else {

        if(read_write) {

          write_variables_.insert(var_ref->get_symbol());
          read_variables_.insert(var_ref->get_symbol());
          return;
        }

        if(left_of_assignment) {
          write_variables_.insert(var_ref->get_symbol());
        } else {
          read_variables_.insert(var_ref->get_symbol());
        }
      }
    }
  }

  if(isSgFunctionCallExp(expression)) {

    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expression);
    SgExpressionPtrList expr_list
      = func_call_exp->get_args()->get_expressions();

    // This flag is important for handling references. If a function call
    // argument is passed by reference, we have to decide if it is a read/write
    // conflict. In case that the function is already analyzed, we know that we
    // solve the reference later. So, we don't set it as a canonical read/write
    // conflict
    bool is_precached = false;

    // Here we have to check if any parameter type is a pointer.
    // If so, the segment is with all other segments in conflict
    SgFunctionDeclaration *func_decl
      = tools::get_first_non_defining_function_declaration(func_call_exp);

    if(func_decl) {

      SgFunctionDeclaration *func_decl_def
        = isSgFunctionDeclaration(func_decl->get_definingDeclaration());
      if(func_decl_def) {
        SgFunctionDefinition *func_def = func_decl_def->get_definition();

        if(func_def) {
          sg::CachedFunctionAstAttributes *precached =
            dynamic_cast<CachedFunctionAstAttributes*>(
                func_def->getAttribute("precached_info"));

          if(precached) {
            is_precached = true;
          }
        }
      }

      // TODO This should not be in the if statement. In the case that we have
      //      virtual function, this part is not analyzed
      SgFunctionParameterList *param_list = func_decl->get_parameterList();
      SgInitializedNamePtrList &init_name_list = param_list->get_args();

      int parameter_counter = 0;
      for(SgInitializedNamePtrList::iterator
          iter  = init_name_list.begin();
          iter != init_name_list.end();
          iter++, parameter_counter++) {

        if(isSgPointerType((*iter)->get_type())) {

          if(parameter_counter < expr_list.size()) {

            if(tools::has_no_variable_expressions(
                  expr_list.at(parameter_counter)) == false) {

              if(debugging_) {

                Conflict conflict(expression);
                conflicts_.insert(conflict);
              }
              in_conflict_with_all_other_segments = true;
            }
          } else {
            if(debugging_) {

              Conflict conflict(expression);
              conflicts_.insert(conflict);
            }
            in_conflict_with_all_other_segments = true;
          }
        }
      }
    }

    int argument_counter = 0;
    SgExprListExp *arguments = func_call_exp->get_args();
    for(SgExpressionPtrList::iterator
        expr_iter  = expr_list.begin();
        expr_iter != expr_list.end();
        expr_iter++, argument_counter++) {

      if(func_decl) {
        if(argument_counter
            < func_decl->get_parameterList()->get_args().size()) {
          if(isSgReferenceType(func_decl->get_parameterList()->get_args().
                at(argument_counter)->get_type())) {

            analyze_expression(*expr_iter, false, false, false, true,
                is_precached, tools::get_most_outer_var(*expr_iter));
          } else {

            analyze_expression(*expr_iter, false, false, false, false, false,
                NULL);
          }
        } else {

          analyze_expression(*expr_iter, false, false, false, false, false,
              NULL);
        }
      }
    }
  }

  if(isSgPntrArrRefExp(expression)) {
    SgPntrArrRefExp *pointer_array_ref = isSgPntrArrRefExp(expression);

    analyze_expression(pointer_array_ref->get_lhs_operand(),
        left_of_assignment,
        read_write,
        member_access,
        function_parameter,
        is_precached_function,
        outer_variable);
    analyze_expression(pointer_array_ref->get_rhs_operand(), false, false,
        false, function_parameter, is_precached_function, outer_variable);
  }

  if(isSgAssignOp(expression)) {
    SgAssignOp *assign_op = isSgAssignOp(expression);
    analyze_expression(assign_op->get_lhs_operand(), true, false, false,
        function_parameter, is_precached_function, outer_variable);
    analyze_expression(assign_op->get_rhs_operand(), false, false, false,
        function_parameter, is_precached_function, outer_variable);
  }

  if(isSgConditionalExp(expression)) {
    SgConditionalExp *conditional_exp = isSgConditionalExp(expression);
    analyze_expression(
        conditional_exp->get_conditional_exp(),
        left_of_assignment,
        read_write,
        false, function_parameter, is_precached_function, outer_variable);
    analyze_expression(conditional_exp->get_true_exp(),
        left_of_assignment,
        read_write,
        false, function_parameter, is_precached_function, outer_variable);
    analyze_expression(conditional_exp->get_false_exp(),
        left_of_assignment,
        read_write,
        false, function_parameter, is_precached_function, outer_variable);
  }

  if(isSgPointerDerefExp(expression)) {
    SgPointerDerefExp *op = isSgPointerDerefExp(expression);
    analyze_expression(op->get_operand(), false, false, false,
        function_parameter, is_precached_function, outer_variable);
    in_conflict_with_all_other_segments = true;
    if(debugging_) {
      Conflict conflict(expression);
      conflicts_.insert(conflict);
    }

    return;
  }

  if(isSgAddressOfOp(expression)) {
    SgAddressOfOp *op = isSgAddressOfOp(expression);
    analyze_expression(op->get_operand(), false, false, false,
        function_parameter, is_precached_function, outer_variable);
    in_conflict_with_all_other_segments = true;
    if(debugging_) {

      Conflict conflict(expression);
      conflicts_.insert(conflict);
    }

    return;
  }

  assert(true && "Operator not supported");
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
