#include "command_line_parsing.h"

#include <limits.h>
#include <cstring>

#include "rose.h"
#include "commandline_processing.h"

#include "../macros.h"
#include "../internal_representation/design.h"

std::vector<std::string>
risc::tools::preparsing_commandline_arguments(int &argc, char **&argv,
    std::vector<std::string> &input_files)
{
  std::vector<std::string> arguments =
    CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

  input_files = CommandlineProcessing::generateSourceFilenames(arguments, true);

  if(input_files.size() == 0)
    return arguments;

  argc = 0;
  argv = NULL;

  std::string new_file("__main__.cpp");

  std::ofstream myfile;
  myfile.open(new_file.c_str(), std::ios::out);

  if(!myfile.is_open()) {
    std::cout << "Cannot create merged file" << std::endl;
    return arguments;
  }

  for(std::vector<std::string>::iterator
      iter  = input_files.begin();
      iter != input_files.end();
      iter++) {

    if(CommandlineProcessing::isSourceFilename(*iter)) {

      myfile << "#include \"" << (*iter).c_str() << "\"" << std::endl;

      arguments.erase(
                    std::remove(arguments.begin(), arguments.end(), *iter),
                    arguments.end());
    }
  }

  myfile.close();

  arguments.insert(++(arguments.begin()), new_file);

  CommandlineProcessing::generateArgcArgvFromList(arguments,argc,argv);

  return arguments;
}

std::string
risc::tools::instrumentor_output_filename(int &argc, char **&argv)
{
  for(int i=0; i < argc-1; i++) {

    if(strcmp("-inst:output_filename", argv[i]) == 0) {
      std::string filename(argv[i+1]);

      // Here we are deleting the arguments '-inst:output_filename'
      std::vector<std::string> arguments =
        CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

      char **temp = NULL;

      arguments.erase(arguments.begin() + i, arguments.begin() + (i + 2));
      CommandlineProcessing::generateArgcArgvFromList(arguments,argc,temp);
      argv = temp;

      return filename;
    }
  }

  // in this case no filename is given, so we use the default name
  std::string filename("__temp__dynamic_analysis.dir");

  return filename;
}

std::string
risc::tools::dynamic_analysis_input_filename(int &argc, char **&argv)
{
  for(int i=0; i < argc-1; i++) {

    if(strcmp("-risc:elab", argv[i]) == 0) {
      std::string filename(argv[i+1]);

      // Here we are deleting the arguments '-dyn:input_filename'
      std::vector<std::string> arguments =
        CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

      char **temp = NULL;

      arguments.erase(arguments.begin() + i, arguments.begin() + (i + 2));
      CommandlineProcessing::generateArgcArgvFromList(arguments,argc,temp);
      argv = temp;

      return filename;
    }
  }

  std::string filename;
  return filename;
}

risc::tools::ParseStatus::ParseStatus()
{
  verbosityLevel = 0;
  warningLevel = 0;
  compilerName = "g++";
  mic = false;
}

risc::tools::ParseStatus& 
risc::tools::ParseStatus::operator=(const risc::tools::ParseStatus& copy)
{
  verbosityLevel = copy.verbosityLevel;
  warningLevel = copy.warningLevel;
  compilerName = copy.compilerName;
  mic = copy.mic;

  return *this;
}

void
risc::tools::ParseCommandlineOptions(int &argc, char** &argv, 
    risc::tools::ParseStatus* status)
{
  Rose_STL_Container<std::string> argList;
  argList = CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

  std::vector<Rose_STL_Container<std::string>::iterator> RemovedIter;

  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    // -v, --verbose
    if (it->compare("-v") == 0 || it->compare("--verbose") == 0)
    {
      status->verbosityLevel ++;
      RemovedIter.push_back(it);
    }

    // -vv
    if (it->compare("-vv") == 0)
    {
      status->verbosityLevel += 2;
      RemovedIter.push_back(it);
    }

    // -vvv
    if (it->compare("-vvv") == 0)
    {
      status->verbosityLevel += 3;
      RemovedIter.push_back(it);
    }

    // -w, --warnings
    if (it->compare("-w") == 0 || it->compare("--warnings") == 0)
    {
      status->warningLevel ++;
      RemovedIter.push_back(it);
    }

    // -ww
    if (it->compare("-ww") == 0)
    {
      status->warningLevel += 2;
      RemovedIter.push_back(it);
    }

    // -www
    if (it->compare("-www") == 0)
    {
      status->warningLevel += 3;
      RemovedIter.push_back(it);
    }
  }

  while (RemovedIter.size() != 0)
  {
    argList.erase(RemovedIter.back());
    RemovedIter.pop_back();
  }

  argc = 0;
  argv = NULL;

  CommandlineProcessing::generateArgcArgvFromList(argList, argc, argv);
}

void
risc::tools::AddCommandlineOption(int &argc, char** &argv, std::string option)
{
  Rose_STL_Container<std::string> argList;
  argList = CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);
  argList.push_back(option);

  argc = 0;
  argv = NULL;

  CommandlineProcessing::generateArgcArgvFromList(argList, argc, argv);
}

void
risc::tools::AddCommandlineOption(risc::Design &design, std::string option)
{
  SgProject* project = design.project_;

  Rose_STL_Container<std::string> argList 
    = project->get_originalCommandLineArgumentList();

  argList.push_back(option);
  project->set_originalCommandLineArgumentList(argList);

  for (int i = 0; i < project->numberOfFiles(); i++)
  {
    SgFile& file = project->get_file(i);
    Rose_STL_Container<std::string> argListFile 
      = file.get_originalCommandLineArgumentList();
    argListFile.push_back(option);
    file.set_originalCommandLineArgumentList(argListFile);
  }
}

void
risc::tools::RemoveCommandlineOption(risc::Design &design, std::string option)
{
  SgProject* project = design.project_;

  std::vector<Rose_STL_Container<std::string>::iterator> RemovedIter;

  // modify the command-line options in SgProject
  Rose_STL_Container<std::string> argList 
    = project->get_originalCommandLineArgumentList();
  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    if (it->compare(option) == 0)
      RemovedIter.push_back(it);
  }
  while (RemovedIter.size() != 0)
  {
    argList.erase(RemovedIter.back());
    RemovedIter.pop_back();
  }
  project->set_originalCommandLineArgumentList(argList);

  // modify the command-line options in SgFile
  for (int i = 0; i < project->numberOfFiles(); i++)
  {
    SgFile& file = project->get_file(i);
    Rose_STL_Container<std::string> argListFile 
      = file.get_originalCommandLineArgumentList();
    for (Rose_STL_Container<std::string>::iterator it = argListFile.begin();
         it != argListFile.end(); it++)
    {
      if (it->compare(option) == 0)
        RemovedIter.push_back(it);
    }
    while (RemovedIter.size() != 0)
    {
      argListFile.erase(RemovedIter.back());
      RemovedIter.pop_back();
    }
    file.set_originalCommandLineArgumentList(argListFile);
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
