#ifndef FUNCTION_HELPER_H_INCLUDED_
#define FUNCTION_HELPER_H_INCLUDED_

#include "rose.h"

#include "../segment_graph/segment_graph.h"

#include <list>

namespace risc
{

namespace tools
{

/**
 * \fn       get_function_declaration_of_channel_call
 * \brief    This function determines which virtual function will be called
 *           in a case of channel access. For instance,
 *           my_port->send(some_data); Through the port declaration and the
 *           function signature we can determine which function will be called.
 *           This concept cannot be generalized.
 *
 * \return   Pointer to the function declaration.
 */
SgFunctionDeclaration*
get_function_declaration_of_channel_call_statically(
    SgFunctionCallExp *func_call);

/**
 * \fn       get_function_declaration_of_channel_call_dynamically
 * \brief    This function determines which function will be called in a case
 *           of channel access. The results from dynamic analysis will be taken
 *           into account.
 * \return   Pointer to the function declaration.
 */
SgFunctionDeclaration*
get_function_declaration_of_channel_call_dynamically(
    SgFunctionCallExp *func_call);

/**
 * \fn     function_call_belongs_to_a_port
 * \brief  This functions checks if the function call belongs to a port.
 *         This function should be only called after a function call is
 *         decomposed through the segment graph.
 * \return True, if the function call belongs to a port
 */
bool
function_call_belongs_to_a_port(SgFunctionCallExp *func_call_exp);

/**
 * \fn    function_call_is_in_user_domain
 * \brief This function checks if a function call is in the user domain
 */
bool
function_call_is_in_user_domain(SgFunctionCallExp *func_call_exp);

/**
 * TODO CHANGE THE NAME
 * \fn    get_all_other_function_declarations
 * \brief This function gets all other functions declarations which are
 *        associated with the same function declaration like func_decl
 */
std::vector<SgFunctionDeclaration*>
get_all_other_function_declarations(SgFunctionDeclaration *func_decl);

/**
 * \fn     get_channel_interfaces
 * \brief  This function returns all used interfaces of the channel channel_def
 */
std::list<SgClassDefinition*>
get_channel_interfaces(SgClassDefinition *channel_def);

/**
 * \fn    get_all_interfaces_in_the_design
 * \brief This function traverses over all channel definitions in the design.
 *        For each individual channel, the function collects all used
 *        interfaces.
 *        This function will not return an interface which is not used in any
 *        channel
 * \param only_user_domain considers only interfaces and channels which are
 *                         defined in the user file
 */
std::list<SgClassDefinition*>
get_all_used_interfaces(bool only_in_user_domain = true);

/**
 * \fn    get_all_port_calls_to_annotated_functions
 * \brief This function returns all port calls
 */
std::list<SgFunctionCallExp*>
get_all_port_calls_to_annotated_functions(risc::sg::SegmentGraph &graph);

/**
 * \fn    get_interface_functions
 * \brief This function returns all function declarations of the channel
 *        channel_def which belong to an interface
 */
std::list<SgFunctionDeclaration*>
get_interface_functions(
    SgClassDefinition *channel_def,
    bool only_annotated = false);

/**
 * \fn    get_all_implemented_interface_functions
 * \brief This functions traverses over all channels and collects the
 *        function declarations of the implemented interfaces
 */
std::list<SgFunctionDeclaration*>
get_all_implemented_interface_functions(
    bool only_in_user_domain = true,
    bool only_annotated = false);

/**
 * \fn    is_annotated_function
 * \brief This function returns true if the annotated
 */
bool is_annotated_function(SgFunctionDeclaration *func_decl);

/**
 * \fn    has_reference_parameter
 * \brief This function returns true if the function has at least one parameter
 *        which is taken via reference
 */
bool has_reference_parameter(SgFunctionDeclaration *func_decl);

/**
 * \fn    get_first_non_defining_function_declaration
 * \brief This function returns for a function call the first non defining
 *        function declaration. It will be considered if it is a port call.
 *        If so, statically or dynamically the port mapping will be solved.
 */
SgFunctionDeclaration*
get_first_non_defining_function_declaration(SgFunctionCallExp *func_call_exp);

}; // end namespace tools

}; // end namespace risc

#endif /* FUNCTION_HELPER_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
